from web3.contract import Contract


class ContractInit:
    def __init__(self, web3) -> None:
        self.web3 = web3

    def __call__(self, address: str, *args, **kwargs) -> Contract:
        return self.web3.eth.contract(
            address=self.web3.toChecksumAddress(address),
            *args,
            **kwargs,
        )

    def calculateAddress(self, address: str, nonce: int = None) -> str:
        """
        Calculates the next address a contract will have based on the
        account deploying it and its nonce
        """
        address = self.web3.toChecksumAddress(address)

        if nonce is None:
            nonce = self.web3.eth.getTransactionCount(address)

        return self.web3.keccak(rlp.encode([bytes.fromhex(address[2:]), nonce]))[
            12:
        ].hex()



import code
import atexit
import os
import readline
import rlcompleter
import pprint

from web3.datastructures import AttributeDict
from hitomi.commands.accounts import Accounts
from hitomi.commands.checksum import Checksum
from hitomi.commands.contract import ContractInit
from hitomi.commands.erc20 import ERC20
from hitomi.commands.abi import Abi
from hitomi.commands.uniswap import UniswapPair
from hitomi.network.web3 import Web3


class Console(code.InteractiveConsole):
    def __init__(self, vars: dict, web3: Web3):
        # Set history file
        history_path = os.path.join(os.environ["HOME"], ".hitomi_history")
        if os.path.isfile(history_path):
            readline.read_history_file(history_path)

        # Trigger history save on exit
        atexit.register(readline.write_history_file, history_path)

        # Setup autocomplete
        readline.set_completer(rlcompleter.Completer(vars).complete)
        readline.parse_and_bind("tab: complete")

        # Load commands
        local_commands = dict(
            {
                "accounts": Accounts(web3),
                "checksum": Checksum(web3),
                "contract": ContractInit(web3),
                "erc20": ERC20(web3),
                "uniswap_pair": UniswapPair(web3),
                "abi": Abi(),
            }
        )
        vars.update(local_commands)

        super().__init__(vars)

    def push(self, line):
        return super().push(line)

    def runsource(self, source, filename="<input>", symbol="single"):
        try:
            code = self.compile(source, filename, "single")
        except (OverflowError, SyntaxError, ValueError):
            self.showsyntaxerror(filename)
            return False

        # Multiline code
        if code is None:
            return True

        try:
            self.compile(source, filename, "eval")
            code = self.compile(f"__ret_value__ = {source}", filename, "exec")
        except Exception:
            pass

        self.runcode(code)

        if "__ret_value__" in self.locals and self.locals["__ret_value__"] is not None:
            self._console_write(self.locals["__ret_value__"])
            del self.locals["__ret_value__"]

        return False

    def _console_write(self, obj):
        """
        Pretty print console output
        """
        text = pprint.pformat(obj, indent=4)
        try:
            if obj and isinstance(obj, AttributeDict):
                objCopy = dict(obj)

                if objCopy.get("logs"):
                    logs = []
                    for log in objCopy.get("logs"):
                        logs.append(dict(log))
                    objCopy["logs"] = logs

                text = pprint.pformat((objCopy), indent=4)
        except Exception as e:
            print(e)

        print(text)

    def _recursive_pretty_print(self, obj):
        try:

            for attr, value in obj.__dict__.items():
                obj = self._recursive_pretty_print()
            pass
        except:

            pass

        return obj



from typing import Any, Dict, Iterator, List, Optional, Tuple, Union

from hitomi.network.web3 import Web3


class Accounts(list):
    def __init__(self, web3) -> None:
        self.web3 = web3
        self.refresh()

    def refresh(self):
        self._cache = dict({"accounts": self.web3.eth.accounts})

    def _accounts(self) -> list:
        return self._cache["accounts"]

    def __iter__(self) -> Iterator:
        return iter(self._accounts())

    def __getitem__(self, key: int) -> Any:
        accounts = self._accounts()
        return accounts[key]

    def __delitem__(self, key: int) -> None:
        accounts = self._accounts()
        del accounts[key]

    def __len__(self) -> int:
        return len(self._accounts())



import json
import os


class ERC20:
    abi = None

    def __init__(self, web3) -> None:
        self.web3 = web3
        with open(
            "{dir}/{file}".format(dir=os.path.dirname(__file__), file="abi.json")
        ) as f:
            self.abi = json.load(f)



import json


class Abi:
    def __init__(self):
        pass

    def fromFile(self, filename: str) -> object:
        with open(filename, "r") as f:
            abi_object = json.load(f)

        return abi_object



import os

from web3 import Web3 as _web3
from web3 import IPCProvider, HTTPProvider, WebsocketProvider
from web3.middleware import geth_poa_middleware


class Web3(_web3):
    """
    Web3 wrapper class
    """

    enode = None

    def __init__(self) -> None:
        super().__init__(HTTPProvider("null"))

    def handleEnode(self) -> None:
        """
        Obtain enode path
        """
        connected = self.isConnected()

        if connected:
            if self.clientVersion.startswith("Parity"):
                self.enode = self.parity.enode()
            elif self.clientVersion.startswith("Geth"):
                self.enode = self.geth.admin.nodeInfo()["enode"]
            elif self.clientVersion.startswith("bor"):
                self.enode = self.geth.admin.nodeInfo()["enode"]

    def handleProofOfAuthorityChain(self) -> None:
        """
        Add PoA middleware because
        https://web3py.readthedocs.io/en/stable/middleware.html#why-is-geth-poa-middleware-necessary
        """
        connected = self.isConnected()

        if connected:
            if self.clientVersion.startswith("bor"):
                self.middleware_onion.inject(geth_poa_middleware, layer=0)

    def connect(self, node: str, timeout: int = 10) -> None:
        try:
            if os.path.exists(node):
                self.provider = IPCProvider(node)
                return
        except OSError:
            pass

        if node.startswith("https://") or node.startswith("http://"):
            self.provider = HTTPProvider(node, request_kwargs={"timeout": timeout})
        elif node.startswith("ws://") or node.startswith("wss://"):
            self.provider = WebsocketProvider(
                node, websocket_kwargs={"timeout": timeout}
            )
        else:
            raise ValueError(
                "The provided node is not valid. It must start with 'http://' or 'https://' or 'ws://' or a path to an IPC socket file."
            )

        try:
            self.handleEnode()
        except Exception:
            pass

        try:
            self.handleProofOfAuthorityChain()
        except Exception:
            pass
